package terminal;

import static terminal.ControlSequenceHandler.isEscape;
import static util.Characters.EOF;

import java.io.*;
import java.util.concurrent.*;

import com.googlecode.lanterna.TerminalPosition;
import com.googlecode.lanterna.input.*;
import com.googlecode.lanterna.terminal.IOSafeTerminal;

public class LispTerminal {

    public static final char END_OF_SEGMENT = 'x';

    private IOSafeTerminal terminal;
    private PipedOutputStream inputWriter;
    private PipedInputStream outputReader;
    private ExecutorService executorService;
    private ControlSequenceHandler controlSequenceHandler;
    private TerminalPosition origin;
    private TerminalPosition leadingEdge;
    private String inputLine;
    private String outputSegment;
    private boolean isFinished;

    public LispTerminal(IOSafeTerminal terminal, PipedOutputStream inputWriter, PipedInputStream outputReader) {
        // FIXME - add resize handler
        this.terminal = terminal;
        this.inputWriter = inputWriter;
        this.outputReader = outputReader;
        this.executorService = Executors.newFixedThreadPool(2);
        this.controlSequenceHandler = new ControlSequenceHandler();
        this.origin = terminal.getCursorPosition();
        this.leadingEdge = origin;
        this.inputLine = "";
        this.outputSegment = "";
        this.isFinished = false;
    }

    public void run() {
        executorService.execute(this::readInput);
        executorService.execute(this::writeOutput);
        executorService.shutdown();
    }

    public void readInput() {
        while (!isFinished)
            processNextKey();
    }

    private void processNextKey() {
        handleKey(getKeyStroke());
        takeNap();
    }

    private KeyStroke getKeyStroke() {
        KeyStroke keyStroke = null;

        try {
            keyStroke = terminal.pollInput();
        } catch (IllegalStateException e) {
            // TODO - Issue #299
            terminal.putCharacter('\n');
            System.exit(0);
        }

        return keyStroke;
    }

    private synchronized void handleKey(KeyStroke keyStroke) {
        if (keyStroke == null)
            return;

        doKey(keyStroke);
        terminal.flush();
    }

    private synchronized void doKey(KeyStroke keyStroke) {
        KeyType keyType = keyStroke.getKeyType();

        if (keyStroke.isCtrlDown())
            doControlKey(keyStroke);
        else if (keyType == KeyType.ArrowLeft)
            moveCursorLeft();
        else if (keyType == KeyType.ArrowRight)
            moveCursorRight();
        else if (keyType == KeyType.Enter)
            doEnter();
        else if (keyType == KeyType.Backspace)
            doBackspace();
        else if (keyType == KeyType.Delete)
            doDelete();
        else if (keyType == KeyType.Character)
            doCharacter(keyStroke);
    }

    private synchronized void doControlKey(KeyStroke keyStroke) {
        KeyType keyType = keyStroke.getKeyType();

        if (keyType == KeyType.Character)
            doControlCharacter(keyStroke);
    }

    private synchronized void doControlCharacter(KeyStroke keyStroke) {
        if (keyStroke.getCharacter() == 'd')
            doControlD();
    }

    private synchronized void doControlD() {
        doEnter();
        finish();
    }

    private synchronized void moveCursorLeft() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveLeft(cursorPosition))
            retractCursor(cursorPosition);
    }

    private synchronized boolean isPossibleToMoveLeft(TerminalPosition cursorPosition) {
        return distanceFromOrigin(cursorPosition) > 0;
    }

    private synchronized int distanceFromOrigin(TerminalPosition cursorPosition) {
        int columnDifference = cursorPosition.getColumn() - origin.getColumn();
        int rowDifference = cursorPosition.getRow() - origin.getRow();
        int totalColumns = terminal.getTerminalSize().getColumns();

        return columnDifference + (totalColumns * rowDifference);
    }

    private synchronized void retractCursor(TerminalPosition cursorPosition) {
        TerminalPosition newPosition = cursorPosition.withRelativeColumn(-1);

        if (isAtStartOfRow(cursorPosition))
            newPosition = cursorPosition.withColumn(terminal.getTerminalSize().getColumns()).withRelativeRow(-1);

        terminal.setCursorPosition(newPosition);

    }

    private boolean isAtStartOfRow(TerminalPosition cursorPosition) {
        return cursorPosition.getColumn() == 0;
    }

    private synchronized void moveCursorRight() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveRight(cursorPosition))
            advanceCursor(cursorPosition);
    }

    private synchronized boolean isPossibleToMoveRight(TerminalPosition cursorPosition) {
        return distanceFromOrigin(cursorPosition) < inputLine.length();
    }

    private synchronized void advanceCursor(TerminalPosition cursorPosition) {
        if (isAtEndOfRow(cursorPosition))
            moveCursorToNextRow(cursorPosition);
        else
            terminal.setCursorPosition(cursorPosition.withRelativeColumn(1));

        TerminalPosition newPosition = terminal.getCursorPosition();

        // FIXME - use distance from origin to set leading edge
        if (distanceFromOrigin(newPosition) == inputLine.length())
            leadingEdge = terminal.getCursorPosition();
    }

    private synchronized boolean isAtEndOfRow(TerminalPosition cursorPosition) {
        return cursorPosition.getColumn() >= terminal.getTerminalSize().getColumns() - 1;
    }

    private void moveCursorToNextRow(TerminalPosition cursorPosition) {
        if (isEndOfBuffer())
            createNewRowForCursor();
        else
            terminal.setCursorPosition(cursorPosition.withColumn(0).withRelativeRow(1));
    }

    private boolean isEndOfBuffer() {
        return terminal.getCursorPosition().getRow() == terminal.getTerminalSize().getRows() - 1;
    }

    private void createNewRowForCursor() {
        TerminalPosition originalPosition = terminal.getCursorPosition();
        terminal.putCharacter('\n');
        terminal.setCursorPosition(originalPosition.withColumn(0));
        origin = origin.withRelativeRow(-1);
    }

    private synchronized void doEnter() {
        moveCursorToEndOfInput();
        terminal.putCharacter('\n');
        inputLine += "\n";

        try {
            inputWriter.write(inputLine.getBytes());
            inputWriter.flush();
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        inputLine = "";
        origin = terminal.getCursorPosition();
        leadingEdge = origin;
    }

    private synchronized void doBackspace() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveLeft(cursorPosition)) {
            String remaining = inputLine.substring(distanceFromOrigin(cursorPosition), inputLine.length());
            inputLine = inputLine.substring(0, distanceFromOrigin(cursorPosition) - 1) + remaining;

            retractCursor(cursorPosition);

            for (char c : remaining.toCharArray())
                terminal.putCharacter(c);

            terminal.putCharacter(' ');

            retractCursor(cursorPosition);

            // FIXME - use distance from origin to set leading edge
            if (distanceFromOrigin(leadingEdge) > inputLine.length())
                leadingEdge = terminal.getCursorPosition();
        }
    }

    private synchronized void doDelete() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveRight(cursorPosition)) {
            String remaining = inputLine.substring(distanceFromOrigin(cursorPosition) + 1, inputLine.length());
            inputLine = inputLine.substring(0, distanceFromOrigin(cursorPosition)) + remaining;

            for (char c : remaining.toCharArray())
                terminal.putCharacter(c);

            terminal.putCharacter(' ');
            terminal.setCursorPosition(cursorPosition);

            // FIXME - use distance from origin to set leading edge
            if (distanceFromOrigin(leadingEdge) > inputLine.length())
                leadingEdge = terminal.getCursorPosition();
        }
    }

    private synchronized void doCharacter(KeyStroke keyStroke) {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveRight(cursorPosition)) {
            String remaining = keyStroke.getCharacter()
                               + inputLine.substring(distanceFromOrigin(cursorPosition), inputLine.length());
            inputLine = inputLine.substring(0, distanceFromOrigin(cursorPosition)) + remaining;

            // FIXME - must have a way to push remainder on to a new line at the end of the buffer

            for (char c : remaining.toCharArray())
                terminal.putCharacter(c);

            advanceCursor(cursorPosition);
        } else {
            terminal.putCharacter(keyStroke.getCharacter());
            inputLine += keyStroke.getCharacter();

            advanceCursor(cursorPosition);
        }
    }

    private void takeNap() {
        try {
            Thread.sleep(1);
        } catch (InterruptedException e) {
            isFinished = true;
        }
    }

    public void writeOutput() {
        try {
            for (int c = outputReader.read(); c != EOF; c = outputReader.read())
                processOutput((char) c);
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        terminal.setCursorVisible(true);
        terminal.flush();
        terminal.close();
    }

    private synchronized void processOutput(char c) {
        if (isEscape(c))
            parseControlSequence();
        else if (isEndOfSegment(c))
            printSegment();
        else
            outputSegment += c;
    }

    private synchronized void parseControlSequence() {}

    private synchronized boolean isEndOfSegment(char c) {
        return c == END_OF_SEGMENT;
    }

    private synchronized void printSegment() {
        terminal.setCursorVisible(false);
        moveCursorToEndOfInput();

        for (char c : outputSegment.toCharArray())
            terminal.putCharacter(c);

        moveCursorToNextRowIfNecessary();
        terminal.flush();
        terminal.setCursorVisible(true);
        outputSegment = "";
        origin = terminal.getCursorPosition();
        leadingEdge = origin;
    }

    private synchronized void moveCursorToEndOfInput() {
        terminal.setCursorPosition(leadingEdge);
    }

    private synchronized void moveCursorToNextRowIfNecessary() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isAtEndOfRow(cursorPosition))
            moveCursorToNextRow(cursorPosition);
    }

    public void finish() {
        isFinished = true;

        try {
            inputWriter.close();
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

}
